# Copyright (c) 2022 DomainTools LLC
# Copyright (c) 2012-2015, 2018-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import dnstable
import time

import dnstli2.abort

class Reader(object):
    def __init__(self, fname, iszone=False, isdnssec=False):
        self.fname = fname
        self.robj = dnstable.reader(str(fname), iszone)
        self.iszone = iszone
        self.isdnssec = isdnssec

    def reload(self):
        self.robj.reload()

def is_dnssec(rrtype):
    if rrtype:
        return rrtype.lower() in ('any-dnssec', 'ds', 'rrsig', 'nsec',
                                  'dnskey', 'nsec3', 'nsec3param',
                                  'dlv', 'cds', 'cdnskey', 'ta',
                                  'type43', 'type46', 'type47',
                                  'type48', 'type50', 'type51',
                                  'type59', 'type60', 'type32768', 'type32769')
    else:
        return False

Heartbeat = object()

class Client:
    def __init__(self, conf):
        self.readers = []
        self._heartbeat_interval = conf.get('heartbeat_interval')

        for c in conf['filesets']:
            reader = Reader(c['setfile'], c['iszone'], c['isdnssec'])
            self.readers.append(reader)

    class LookupIter:
        def __init__(self, parent, qobj, limit, heartbeat_interval=None):
            self._parent = parent
            self._qobj = qobj
            self._limit = limit
            self._heartbeat_interval = heartbeat_interval

            self._count = 0
            self._readers = iter(self._parent.readers)
            self._query = iter(tuple())

        def __iter__(self):
            return self

        def __next__(self):
            next_heartbeat = None
            if self._heartbeat_interval:
                next_heartbeat = time.time() + self._heartbeat_interval

            if self._count >= self._limit:
                raise StopIteration

            while True:
                try:
                    rval = next(self._query)
                    self._count += 1
                    return rval
                except dnstable.Timeout:
                    # the dnstable library periodically raises Timeout as it is processing a query.
                    # check_connection() checks if dnstli's calling client is still connected.  If
                    # not, then it raises an Abort here (Note this is not the only path to an
                    # early_exit).  Otherwise, check_connection() returns normally and this while
                    # loop will continue onto the next iteration.
                    dnstli2.abort.check_connection()

                    if self._heartbeat_interval and time.time() >= next_heartbeat:
                        return Heartbeat
                except StopIteration:
                    self._next_reader()

        def _next_reader(self):
            while True:
                reader = next(self._readers)
                # run a dnssec query on a dnssec reader
                # or run a non-dnssec query on a non-dnssec reader
                if is_dnssec(self._qobj.rrtype) == reader.isdnssec:
                    self._query = reader.robj.query(self._qobj)
                    return

    def lookup(self, qobj, limit):
        return Client.LookupIter(self, qobj, limit, heartbeat_interval=self._heartbeat_interval)
