# Copyright (c) 2022 DomainTools LLC
# Copyright (c) 2012-2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

try:
    import redis.exceptions
except:
    pass
from collections import namedtuple
import time
from dnstli2.log import *

# A *RateParams named tuple is passed into the rate methods.
#
# Elements of both tuples:
#
#   authtype =
#     Type of authorization, 'ui' or 'api'
#
#   identifier =
#     A username or group_name. The pair of authtype and identifier can identify the proper
#     ERP subscription detail table name and row.
#
# Elements of AutoQRateParams tuples:
#
#   maximum =
#      Integer or None; The maximum number of API lookups that may be performed during the quota period.
#      Two values of maximum indicate the account is unlimited, None and -1.  Better to use -1.
#      The calling code will not create a AutoQRateParams for a user quota with a maximum=None -- if
#      that code finds dnstli.limit is None then it substitutes it with the dnstli
#      configuration value query_limit_default.
#      maximum must be None for block-based quotas.
#
#   period =
#      Integer or None; Seconds for how often this quota's limit will be reset.
#      Typically, this is 86400, which is one day.
#      If this is None then this is a block-based quota (and blkq_split must be an integer)
#
# Elements of BlkQRateParams tuples:
#
#   blkq_split =
#      Integer or None; The number of API lookups split for block-based quota.
#      This may _never_ decrease.  If this is None then this is *not* a Block-based quota.
#
#   blkq_baseline_used =
#      Integer or None; The number of API lookups used for block-based quota, as a baseline.
#      This may _never_ decrease.  If this is None then this is *not* a Block-based quota.
#
#   blkq_expires =
#      Integer or None; UNIX epoch timestamp with second granularity indicating when the block expires.
#      Ignored if this is *not* a Block-based quota.
#
AutoQRateParams = namedtuple('AutoRateParams',
                             ['authtype', 'identifier', 'maximum', 'period'])

allNoneAutoQRateParams = AutoQRateParams(authtype = None, identifier = None,
                                         maximum = None, period = None)

BlkQRateParams = namedtuple('BlkQRateParams',
                            ['authtype', 'identifier', 'blkq_split', 'blkq_baseline_used', 'blkq_expires',
                             'burst_size', 'burst_window'])

allNoneBlkQRateParams = BlkQRateParams(authtype = None, identifier = None,
                                       blkq_split = None, blkq_baseline_used = None, blkq_expires = None,
                                       burst_size = None, burst_window = None)

# A LimitInfo named tuple is returned by the rate methods.
#
# It contains one of two types of limits.
#  1. Auto-replenishing time-based quotas which use parameters: reset, limit, and remaining.
#  2. Block-based quotas which use those parameters plus expires.
#
# Elements of the tuple:
#
#   reset =
#      UNIX epoch timestamp with second granularity indicating the next point in time when the quota limit will be
#      reset. Usually this is at 00:00 (midnight) UTC.  "n/a" if block-based
#
#   limit =
#      The maximum number of API lookups that may be performed during the quota period.
#
#   remaining =
#      For non-block-based quotas, the number of remaining API lookups that may
#      be performed during the quota period.  Otherwise, the number of API
#      lookups split for block-based quota minus the number of API lookups used.
#
#   expires =
#      UNIX epoch timestamp with second granularity indicating when the block expires.  None if not block-based
#
LimitInfo = namedtuple('LimitInfo', ['limit', 'remaining', 'reset', 'expires'])

# Unlimited (auto-replenishing) time-based quota
unlimited = LimitInfo(limit='unlimited', remaining='n/a', reset='n/a', expires=None)

class RedisConnError(Exception):
    pass

class LimitExceeded(Exception):
    pass

class BurstLimitExceeded(Exception):
    pass

class QuotaExpired(Exception):
    pass

class RateLimiter(object):
    def __init__(self, redis_server, default_period):
        self.redis_server = redis_server
        self.default_period = default_period

    def _autoq_get_or_incr_limit(self, rateParams, incr=False):
        '''
        This is an auto-replenishing quota case.
        Expects a list of well-formed AutoQRateParams tuples, or None.
        Uses redis to track how many queries have been made.
        If incr is True then this will increment/get the key in/from the redis cache.
        This will raise LimitExceeded if we exceed the maximum.
        This will raise RedisConnError if we get various Redis errors.
        If incr is False then this will just get the key from the redis cache, for reporting.

        Returns a LimitInfo tuple with possibly-incremented limit.
        '''
        # if there are no rateParams or if all the maximas are -1 or None, then limit is unlimited
        if all(rp.maximum < 0 for rp in rateParams if rp.maximum is not None):
            return unlimited

        t0 = int(time.time())
        reset = None

        remaining = None
        # limit is the smallest maximum, treating a maximum of -1 or None like +infinity
        limit = min(rp.maximum for rp in rateParams if rp.maximum is not None and rp.maximum >= 0)

        for rp in rateParams:
            if rp.maximum is None or rp.maximum < 0:
                continue

            if not rp.identifier:
                continue

            if not rp.maximum:
                raise ValueError('Maximum not defined for {}'.format(rp.identifier))

            period = rp.period
            if period is None:
                period = self.default_period

            if period > 0:
                t1 = t0 - (t0 % period)
                key = 'rate/%s/%s/%d' % (rp.authtype, rp.identifier, t1)
            else:
                t1 = 0
                key = 'rate/%s/%s/fixed' % (rp.authtype, rp.identifier)

            try:
                if not incr:
                    try:
                        current = int(self.redis_server.get(key))
                    except TypeError:
                        current = 0
                else:
                    current = self.redis_server.incr(key)
                    self.redis_server.expire(key, period) # expire key after end of period

                    if current > rp.maximum:
                        raise LimitExceeded
            except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
                raise RedisConnError

            # This is here because we cannot guarantee that someone
            # has not put a negative number into redis, although in
            # practice it will never happen
            if current < 0:
                current = 0

            cur_remaining = rp.maximum - current
            if cur_remaining < 0:
                cur_remaining = 0
            if remaining is None or cur_remaining < remaining:
                remaining = cur_remaining
                if period > 0:
                    reset = t1 + period

        return LimitInfo(limit=limit, remaining=remaining, reset=reset, expires=None)

    def autoq_get_limit(self, rps):
        '''
        Expects a list of RateParams tuples.
        Returns a LimitInfo tuple with unchanged limit
        '''
        return self._autoq_get_or_incr_limit(rps, incr=False)

    def autoq_incr_limit(self, rps):
        '''
        Expects a list of RateParams tuples.
        Returns a LimitInfo tuple with the incremented limit
        This will raise LimitExceeded if we exceed the maximum.
        '''
        return self._autoq_get_or_incr_limit(rps, incr=True)

    def _blkq_get_or_incr_limit(self, rateParams, incr=False):
        '''
        This is a block-quota case
        $$$ Not for MVP: handle groups
        Expects a list of well-formed RateParams tuples.
        Uses redis to track how many queries have been made.
        If incr is True then this will increment/get the key in/from the redis cache.
        This will raise LimitExceeded if we exceed the maximum.
        This will raise BurstLimitExceeded if we exceed the burst rate limit.
        This will raise RedisConnError if we get various Redis errors.
        If incr is False then this will just get the key from the redis cache, for reporting.

        Returns a LimitInfo tuple with possibly-incremented limit.
        '''
        if len(rateParams) != 1:
            raise ValueError('Need one RateParams tuple')

        rp = rateParams[0]      # $$$ only uses first of the rateParams

        # enforce minimum values.  Note that in Python2: None < 0
        if rp.blkq_baseline_used is None:
            tmp_baseline_used = 0
        else:
            tmp_baseline_used = max(rp.blkq_baseline_used, 0)

        tmp_split = max(rp.blkq_split, 0)

        if not rp.identifier:
            raise ValueError('Missing identifier in rateParams %s' % str(rp))

        try:
            ### Begin burst rate quota
            if incr:                # burst windows only looked at if it matters
                b_w = rp.burst_window
                # for python2, can use just "bw > 0" but python3 throws: TypeError: unorderable types: NoneType() > int()
                if b_w is not None and b_w > 0:
                    # Use the last example on https://redis.io/commands/incr (as of 12/26/18) to time
                    # out the redis keys.  Note the race here, but it's not a problem: EXISTS may return false but the
                    # key may be created by another client before we create it inside the pipeline's transaction.  However
                    # this race will just allow an extra query under rare conditions, so the rate limiting will still work
                    # correctly.
                    # Note that this code assumes a "small" burst_size, since it creates a list member for each query.
                    # This should be fine for "small" up to a few thousand.
                    b_key = 'b/' + rp.identifier
                    num_in_b = self.redis_server.llen(b_key)
                    if num_in_b is not None and num_in_b >= rp.burst_size:
                        raise BurstLimitExceeded
                    if self.redis_server.exists(b_key) == False:
                        self.redis_server.pipeline().rpush(b_key, 'x').expire(b_key, b_w).execute()
                    else:
                        self.redis_server.rpushx(b_key, 'x')
            ### End burst rate quota

            key = 'blkq_used/%s/%s' % (rp.authtype, rp.identifier)
            log(DEBUG, "_get_or_incr_limit BLKQ key=%s" % key)

            # Get the current blocks used from redis.
            # If not present, set it to the baseline used value.
            # Careful with the race condition that the key was missing for our GET and for another dnstli's GET,
            # but they beat us to the SET, and then we tried to SET it.  We use the nx=True flag to ensure
            # that only the first writer wins.
            used_str = self.redis_server.get(key)
            if used_str is None:
                # SET if NOT EXISTS
                res = self.redis_server.set(key, tmp_baseline_used, nx=True)
                if res == "OK":
                    log(DEBUG, 'initial SET of key %s' % key)
                else:
                    log(DEBUG, 'someone beat us to initial SET of key %s' % key) # we lost a race but that's ok
                used = tmp_baseline_used
            else:
                try:
                    used = int(used_str)
                except TypeError:
                    raise ValueError('key %s value exists but is not an integer %s' % (key, used_str))

            if incr:
                if int(time.time()) >= rp.blkq_expires:
                    # block quota has expired, a "normal" condition
                    log(DEBUG, 'block quota expired key=%s now=%d expires=%s' % (key, int(time.time()), rp.blkq_expires))
                    raise QuotaExpired

                if used >= tmp_split:
                    log(DEBUG, 'block quota rate limit key=%s used=%s split=%s' % (key, used, tmp_split))
                    raise LimitExceeded

                used = self.redis_server.incr(key)

            # This is here because we cannot guarantee that someone has not put a negative number into redis,
            # although in practice it will never happen
            if used < 0:
                    used = 0

            # Handle rare case: if redis thinks less was used than ERP then override redis to use ERP's value
            if used < tmp_baseline_used:
                log(INFO, 'block quota key %s value %s too low; overriding redis to %s' % (key, used, tmp_baseline_used))
                # SET whether or not it exists
                self.redis_server.set(key, tmp_baseline_used) # There should not be a return value from SET
                used = tmp_baseline_used

        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
            raise RedisConnError

        # If in ERP the quota is reduced, then blkq_split will decrease.  If we already used more than the
        # new quota, (tmp_split - used) would be negative -- in that case, report 0
        return LimitInfo(limit=tmp_split, remaining=max(tmp_split - used, 0), reset="n/a", expires=rp.blkq_expires)

    def blkq_get_limit(self, rps):
        '''
        Expects a list of RateParams tuples.
        Returns a LimitInfo tuple with unchanged limit
        '''
        return self._blkq_get_or_incr_limit(rps, incr=False)

    def blkq_incr_limit(self, rps):
        '''
        Expects a list of RateParams tuples.
        Returns a LimitInfo tuple with the incremented limit
        This will raise LimitExceeded if we exceed the maximum.
        This will raise BurstLimitExceeded if we exceed the burst rate limit.
        '''
        return self._blkq_get_or_incr_limit(rps, incr=True)
