# Copyright (c) 2012-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import distutils.version
import os
import sys
import time

import uwsgi
from uwsgidecorators import postfork, timer

import dnstable
import dnstli2.client
import dnstli2.conf
import dnstli2.log
import dnstli2.wsgi.api
import dnstli2.wsgi.auth
import dnstli2.wsgi.cors

# Load the readers
try:
    config = dnstli2.conf.get_config()
    print('Loading dnstable readers... in pid', os.getpid())
    t0 = time.time()
    client = dnstli2.client.Client(config)
    t1 = time.time()
    print('Finished loading dnstable readers in {:.2f} seconds.'.format(t1 - t0))
except dnstable.DnstableException as de:
    print("dnstli2 FATAL error opening and processing dns fileset: ", str(de))
    sys.exit(1)
except Exception as e:
    print("dnstli2 FATAL error in getting and applying config: ", str(e))

    sys.exit(1)


@postfork
def open_syslog():
    # Open syslog.
    dnstli2.log.openlog('dnstli2')

# Dummy query objects, needed to run dummy queries.
# This detects when filesets are changed and force an immediate reload,
# since a side effect of the dummy query is to force a reload of the fileset.
dummy_query_dns = dnstable.query(dnstable.RRSET, '.', 'ANY')
dummy_query_dnssec = dnstable.query(dnstable.RRSET, '.', 'ANY-DNSSEC')

# Dummy query function that runs every 60 seconds in worker processes.
# Executes a dummy query on both dns and dnssec.
# Reads the first result object, if any, and skips rest.
@timer(60, target="workers")
def run_dummy_query(*args):
    try:
        res = client.lookup(dummy_query_dns, 1)
        n = next(res)
    except StopIteration:
        pass

    try:
        res = client.lookup(dummy_query_dnssec, 1)
        n = next(res)
    except StopIteration:
        pass

    #print "ran dummy query, pid=", os.getpid(), "my worker id=", "worker"+str(uwsgi.worker_id())

# Runs the dummy queries immediately, to prime the cache.
# The previous @timer code will run them also, but after the first timer
# firing..
@postfork
def run_dummy_queries_now():
    run_dummy_query()

# The variables apikey and noauth each reference an anonymous function
# returned from the cors module.  These variables are the "module"
# names, aka WSGI application entry points, i.e., the names of these
# variables are referenced from uWSGI configuration, so they should
# not be changed without coordination.  The variable apikey is only
# created if its preconditions are satisfied.  Exactly one of the
# functions will be called by invoking uwsgi with the corresponding
# "module" name.
if os.path.isdir(dnstli2.conf.API_DIR):
    apikey = dnstli2.wsgi.cors.cors(dnstli2.wsgi.auth.apikey(dnstli2.wsgi.api.DNSTLI(client)))
else:
    uwsgi.log('Directory {!r} does not exist.  Not instantiating apikey'.format(dnstli2.conf.API_DIR))

noauth = dnstli2.wsgi.cors.cors(dnstli2.wsgi.auth.noauth(dnstli2.wsgi.api.DNSTLI(client)))
