# Copyright (c) 2023 DomainTools LLC
# Copyright (c) 2012-2016, 2018-2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import binascii
from collections import OrderedDict
import itertools
from io import StringIO
import json
import time
import traceback
import uuid
import urllib.parse
import re
import sys

import dnstable
import wdns

from dnstli2.log import log, NOTICE, DEBUG
import dnstli2.abort
import dnstli2.client
import dnstli2.conf
import dnstli2.filter
import dnstli2.fmt
import dnstli2.response
import dnstli2.util

conf = dnstli2.conf.get_config()

json_decoder = json.JSONDecoder()

# Truncate swclient, version, and id parameters for logging purposes.
MAX_SWCLIENT_LENGTH = 20
MAX_VERSION_LENGTH = 20
MAX_ID_LENGTH = 30

# Maximum buffered data in output of lookup endpoint
MAX_PAGER_BYTES = 4096


def extractHumanTime(qs):
    ''' Return True/False for the "humantime" flag.  Return False if unclear or not specified.
        Actually returns False unless the string starts with t or T.
    '''
    try:
        flag = qs.get('humantime', None)
        if flag:
            flag = flag[0]
            if len(flag) > 0 and (flag[0] == 't' or flag[0] == 'T'):
                return True
    except:
        pass
    return False

def extractId(qs):
    ''' Return a well-formed user id or return None if anything is wrong or if it is missing '''
    id = qs.get('id')
    if id:
        return id[0][:MAX_ID_LENGTH]
    return None

def extractAggr(qs):
    ''' Return True/False if the query string requests aggregation.  Return True if unclear or not specified.
        Actually returns True unless the string starts with f or F.
    '''
    try:
        aggr = qs.get('aggr', None)
        if aggr:
            aggr = aggr[0]
            if len(aggr) > 0 and (aggr[0] == 'f' or aggr[0] == 'F'):
                return False
    except:
        pass
    return True

def extractOffset(qs):
    ''' Return integer number of rows to offset (e.g. skip) if the query string requests it.  Return 0 if not specified.
    '''
    try:
        offset = qs.get('offset', 0)[0]
        if offset:
            offset = int(offset)
            if offset > 0:
                return offset
    except:
        pass
    return 0

def extractSWClient(qs):
    ''' Return a well-formed "swclient" or return None if anything is wrong or if it is missing.
        As a fallback for the previous deployed version of Scout, if "swclient" is missing, then try "client"
    '''
    # Try "swclient" first, fallback to "client"
    swclient = qs.get('swclient', qs.get('client'))
    if swclient:
        return swclient[0][:MAX_SWCLIENT_LENGTH]
    return None

def extractVersion(qs):
    ''' Return a well-formed user version or return None if anything is wrong or if it is missing '''
    version = qs.get('version')
    if version:
        return version[0][:MAX_VERSION_LENGTH]
    return None

# The following query parameters are already logged separately.
# We want to log all the others.
alreadyLoggedQueryParams = ('swclient', 'id', 'version', 'aggr')

def extractOtherQueryParamsToLog(qs):
    ''' Return other query parameters from a query string that we do not already log in some fashion.
        Return None if there is any error or there are none to log..

        The other parameters will go into a new keyvalue in the "lookup" keyvalue called
        "otherParams", if there are any of them.

        There are some parameters that won't be logged.
        Given "&param=4&param=2&keyonly" only "param": "4" will be logged.
    '''
    try:
        others = dict()
        for qp in qs:
            if qp not in alreadyLoggedQueryParams:
                others[qp] = qs[qp][0]
        return others
    except:
        return None

class QueryTimeExceeded(Exception):
    pass

class QueryTimeExceededIncomplete(Exception):
    pass

class DNSTLI:
    def __init__(self, client):
        self.client = client

    def __call__(self, env, start_response):
        ''' This is called by auth.py code like: return self.app(env, start_response_add_limit_headers)
        '''
        return Lookup(self.client, env, start_response)

class Lookup:
    def __init__(self, client, env, start_response):
        self.client = client
        self.env = env
        self.start_response = start_response
        self.use_epoch_time = False
        self.row_count = 0
        self.max_count = None
        self.limit = 0
        self.bytes = 0
        self.secs = 0
        self.secs_to_first = 0
        self.max_query_time = conf['max_query_time']
        self.offset_max = conf['offset_max']
        self.offset = 0
        self.aggregate = True
        self.querierId = None
        self.querierSWClient = None
        self.querierVersion = None
        self.returnContentType = None

    def __iter__(self):
        try:
            method, args = dnstli2.util.get_method_args(self.env)

            if method == 'lookup':
                self.returnContentType = dnstli2.response.content_type_to_return(self.env,
                                                                                 dnstli2.response.allowed_query_content_types,
                                                                                 dnstli2.response.preferred_query_content_type)

                if args:
                    self.begin_lookup(args)
                    self.start_response(dnstli2.response.Raw200Status, [('Content-Type', self.returnContentType)])
                    yield dnstli2.response.format_saf_row(dnstli2.response.cond_begin).encode('utf-8')

                    for chunk in self.format_response():
                        if self.max_query_time:
                            secs = time.time() - self.t0
                            if secs > self.max_query_time:
                                raise QueryTimeExceededIncomplete
                        self.bytes += len(chunk)
                        yield chunk.encode('utf-8')

                    if self.row_count < self.limit:
                        cond = dnstli2.response.cond_succeeded
                        msg = None
                    else:
                        cond = dnstli2.response.cond_limited
                        msg = "Result limit reached"

                    self.log_lookup()
                    yield dnstli2.response.format_saf_row(cond = cond, msg = msg).encode('utf-8')
                    return
                else:
                    # did not get any query parameters -- that's a bad request
                    raise dnstli2.response.BadRequest

            elif method == 'summarize':
                self.returnContentType = dnstli2.response.content_type_to_return(self.env,
                                                                                 dnstli2.response.allowed_query_content_types,
                                                                                 dnstli2.response.preferred_query_content_type)

                # Note that this ignores the requested content-type,
                # which is how dnstli classically handled unknown content-types
                if args:
                    self.begin_summarize(args)
                    self.start_response(dnstli2.response.Raw200Status, [('Content-Type', self.returnContentType)])

                    # we want to find the earliest *time_first and the latest *time_last
                    earliest_time_first = sys.maxsize
                    latest_time_last = 0
                    earliest_zone_time_first = sys.maxsize
                    latest_zone_time_last = 0
                    total_count = 0
                    num_results = 0

                    yield dnstli2.response.format_saf_row(dnstli2.response.cond_begin).encode('utf-8')
                    next_heartbeat = None
                    if conf.get('heartbeat_interval'):
                        next_heartbeat = time.time() + conf['heartbeat_interval']

                    for entry in self.res:
                        if conf.get('heartbeat_interval') and time.time() >= next_heartbeat:
                            next_heartbeat = time.time() + conf['heartbeat_interval']
                            yield dnstli2.response.format_saf_row().encode('utf-8')

                        if entry is dnstli2.client.Heartbeat:
                            continue

                        self.row_count += 1

                        json_entry = json_decoder.decode(entry.to_json())

                        x_time_first = json_entry.get("time_first", sys.maxsize)
                        x_time_last = json_entry.get("time_last", 0)
                        x_zone_time_first = json_entry.get("zone_time_first", sys.maxsize)
                        x_zone_time_last = json_entry.get("zone_time_last", 0)

                        if x_time_first < earliest_time_first: earliest_time_first = x_time_first
                        if x_time_last > latest_time_last: latest_time_last = x_time_last
                        if x_zone_time_first < earliest_zone_time_first: earliest_zone_time_first = x_zone_time_first
                        if x_zone_time_last > latest_zone_time_last: latest_zone_time_last = x_zone_time_last

                        total_count += json_entry.get("count", 0)
                        num_results += 1

                        if self.max_count != None and total_count >= self.max_count:
                            break

                        if self.max_query_time:
                            secs = time.time() - self.t0
                            if secs > self.max_query_time:
                                raise QueryTimeExceeded

                    # It is ok if no records found (since an actual count from the underlying data
                    # cannot be 0), just return count of 0.
                    summary = '{"count":' + str(total_count) + ',"num_results":' + str(num_results)

                    if earliest_time_first != sys.maxsize:
                        summary += ',"time_first":' + dnstli2.fmt.fmt_time(earliest_time_first,
                                                                              self.use_epoch_time)
                    if latest_time_last != 0:
                        summary += ',"time_last":' + dnstli2.fmt.fmt_time(latest_time_last,
                                                                             self.use_epoch_time)
                    if earliest_zone_time_first != sys.maxsize:
                        summary += ',"zone_time_first":' + dnstli2.fmt.fmt_time(earliest_zone_time_first,
                                                                                   self.use_epoch_time)
                    if latest_zone_time_last != 0:
                        summary += ',"zone_time_last":' + dnstli2.fmt.fmt_time(latest_zone_time_last,
                                                                                  self.use_epoch_time)
                    summary += '}'

                    self.log_lookup(summarize=True)
                    yield dnstli2.response.format_saf_row(obj=summary).encode('utf-8')
                    yield dnstli2.response.format_saf_row(cond = dnstli2.response.cond_succeeded).encode('utf-8')

                    return
                else:
                    # did not get any query parameters -- that's a bad request
                    raise dnstli2.response.BadRequest
            else:
                raise dnstli2.response.BadRequest
        except dnstli2.abort.Abort:
            # This case is when the dnstable library yielded a Timeout and we checked for a client
            # connection, but it was dead so we raised abort to kill the query procesing.  There is
            # no client attached to which to report the query failure.
            tombstone = dnstli2.response.format_saf_row(cond = 'failed', msg = "Processing timeout; results may be incomplete")
            log(DEBUG, "tombstone " + repr(tombstone))
            yield tombstone.encode('utf-8')
            self.log_lookup(early_exit=True, early_cause="query_abort")
        except GeneratorExit:
            # This case is when the generator was closed
            self.log_lookup(early_exit=True)
            raise
        except QueryTimeExceeded:
            # This case is when above in __iter__, we detected that the query processing was taking
            # too long, as we processed individual results.  This is a different case than dnstable
            # raising Timeout.
            log(DEBUG, "query time exceeded, > " + str(self.max_query_time) + " seconds, aborting")
            tombstone = dnstli2.response.format_saf_row(cond = 'failed', msg = "Processing timeout")
            log(DEBUG, "tombstone " + repr(tombstone))
            yield tombstone.encode('utf-8')
            self.log_lookup(early_exit=True, early_cause="query_time_exceeded")
            return
        except QueryTimeExceededIncomplete:
            # Same as QueryTimeExceeded except for the error message
            log(DEBUG, "query time exceeded, > " + str(self.max_query_time) + " seconds, aborting")
            tombstone = dnstli2.response.format_saf_row(cond = 'failed', msg = "Processing timeout; results may be incomplete")
            log(DEBUG, "tombstone " + repr(tombstone))
            yield tombstone.encode('utf-8')
            self.log_lookup(early_exit=True, early_cause="query_time_exceeded")
            return
        except dnstli2.response.HTTPResponse as err:
            self.log_lookup_error(err.status)
            self.start_response(err.status, err.headers)
            if err.message is not None:
                yield err.message
        except:
            status = '500 Internal Server Error'
            headers = [('Content-Type', 'text/html')]

            tb = traceback.format_exc()
            logid = str(uuid.uuid4())
            self.log_lookup_error(status, tb, logid)
            self.start_response(status, headers)
            yield dnstli2.fmt.format_http500(logid).encode('utf-8')

    def log_lookup_error(self, status, tb=None, logid=None):
        msg = dnstli2.util.form_basic_log(self.env, 'lookup-error')
        msg['status'] = status
        if logid:
            msg['logid'] = logid
        if tb:
            msg['traceback'] = json.dumps(tb) # escape embedded quotes etc
        msg['lookup'] = { 'server': 'dnstli2', 'error': status}

        log(NOTICE, json.dumps(msg))

    def log_lookup(self, early_exit=False, early_cause=None, summarize=False):
        if summarize:
            logtype = 'summarize'
        else:
            logtype = 'lookup'

        msg = dnstli2.util.form_basic_log(self.env, logtype)
        msg['count'] = self.row_count
        msg['bytes'] = self.bytes
        msg['time'] = round(self.secs, 3)
        msg['ttfb'] = round(self.secs_to_first, 3)

        if self.querierId:
            msg['clientId'] = self.querierId
        if self.querierSWClient:
            msg['swclient'] = self.querierSWClient
        if self.querierVersion:
            msg['version'] = self.querierVersion

        if early_exit:
            msg['early_exit'] = True

        if early_cause:
            msg['early_cause'] = early_cause

        query = getattr(self, 'query', None)
        if query is not None:
            msg['lookup'] = query.log
        else:
            msg['lookup'] = { 'server' : 'dnstli2' } # don't know a use case where this is called, but just in case

        log(NOTICE, json.dumps(msg))

    def begin_lookup(self, args):
        ''' Returns True if there is at least one response, else returns False or raises an error
        '''

        qs = urllib.parse.parse_qs(self.env.get('QUERY_STRING',''))

        time_filter = dnstli2.filter.TimeFilter()
        try:
            time_filter.apply_query_string(qs)
        except ValueError:
            raise dnstli2.response.BadRequest
        time_filter.apply_policy(self.env.get('dnstli.policy', None))

        # There are four controls for the limit on the number of results to return.  In ascending priority:
        # - global dnstli config default, res_limit_default, used if query doesn't specify a limit and is not reduced
        # - limit query parameter, limit, may be reduced by next two
        # - global dnstli config max, res_limit_max, which will also be used if query limit parameter is 0
        # - per-subscription configuration, dnstli.res_limit_override
        try:
            self.limit = int(qs['limit'][0])
            if self.limit > conf['res_limit_max'] or self.limit == 0:
                self.limit = conf['res_limit_max']
        except ValueError:
            raise dnstli2.response.BadRequest
        except:
            self.limit = conf['res_limit_default']

        num_results_override = self.env.get('dnstli.res_limit_override')
        if num_results_override:
            try:
                self.limit = min(self.limit, int(num_results_override))
            except:
                log(DEBUG, 'Illegal value of res_limit_override "%s" in config' % (num_results_override,))

        self.use_epoch_time = not extractHumanTime(qs)

        self.offset = extractOffset(qs)
        # if offset > 0 specified and got an num_results_override (presumably for a DCE type of restricted power key), fail
        if self.offset > 0 and num_results_override:
            raise dnstli2.response.OffsetDisallowed
        # if offset is too high then fail or fail
        if self.offset_max and self.offset > self.offset_max:
            raise dnstli2.response.OffsetOverMax

        self.aggregate = extractAggr(qs)
        self.querierId = extractId(qs)
        self.querierSWClient = extractSWClient(qs)
        self.querierVersion = extractVersion(qs)

        otherQueryParamsToLog = extractOtherQueryParamsToLog(qs)

        try:
            index, index_type, args = args.split('/', 2)
        except:
            raise dnstli2.response.BadRequest

        try:
            if index == 'rrset':
                self.query = rrsetQuery(index_type, args,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'], aggregate=self.aggregate, offset=self.offset,
                        otherQueryParamsToLog=otherQueryParamsToLog, wildcard=conf['wildcard_label'])
            elif index == 'rdata':
                self.query = rdataQuery(index_type, args,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'], aggregate=self.aggregate, offset=self.offset,
                        otherQueryParamsToLog=otherQueryParamsToLog)
            else:
                raise dnstli2.response.BadRequest

            self.row_count = 0
            self.t0 = time.time()

            # calls client.py:Client.lookup and gets back a Client.LookupIter
            self.res = self.client.lookup(self.query.qobj, self.limit)
        except dnstable.DnstableException:
            # Note that dnstable query timeouts do *not* make it here
            raise dnstli2.response.BadRequest
        return True

    def format_response(self):
        buf = StringIO()

        for entry in self.res:
            if entry is dnstli2.client.Heartbeat:
                # Send a heartbeat message and flush the pager buffer.
                buf.write(dnstli2.response.format_saf_row())
                yield buf.getvalue()
                buf = StringIO()

                continue

            if self.row_count == 0:
                self.secs_to_first = time.time() - self.t0

            self.row_count += 1
            buf.write(dnstli2.response.format_saf_row(obj = self.query.format_entry(entry, self.use_epoch_time)))

            # Flush the pager buffer if we have accumulated enough output.
            if buf.tell() >= MAX_PAGER_BYTES:
                yield buf.getvalue()
                buf = StringIO()
        else:
            yield buf.getvalue()

        self.secs = time.time() - self.t0

    def begin_summarize(self, args):
        qs = urllib.parse.parse_qs(self.env.get('QUERY_STRING',''))

        time_filter = dnstli2.filter.TimeFilter()
        try:
            time_filter.apply_query_string(qs)
        except ValueError:
            raise dnstli2.response.BadRequest
        time_filter.apply_policy(self.env.get('dnstli.policy', None))

        # max_count controls stopping when we reach that summary count.
        # The resulting total count can exceed max_count as it will include the entire count from the last rrset examined.
        try:
            mc = qs.get('max_count')
            if mc:
                self.max_count = int(mc[0])
                if self.max_count <= 0:
                    raise dnstli2.response.BadRequest
        except ValueError:
            raise dnstli2.response.BadRequest

        # There are four controls for the limit on the number of results to count.  In ascending priority:
        # - global dnstli config default, res_limit_default, used if query doesn't specify a limit and is not reduced
        # - limit query parameter, limit, may be reduced by next two
        # - global dnstli config max, res_limit_max, which will also be used if query limit parameter is 0
        # - per-subscription configuration, dnstli.res_limit_override
        #$$$ It might make sense to use a different config file variables
        #$$$ than res_limit_max and res_limit_default to allow
        #$$$ "/summarize" to access more rows than "/lookup".
        try:
            self.limit = int(qs['limit'][0])
            if self.limit > conf['res_limit_max'] or self.limit == 0:
                self.limit = conf['res_limit_max']
        except ValueError:
            raise dnstli2.response.BadRequest
        except:
            self.limit = conf['res_limit_default']

        num_results_override = self.env.get('dnstli.res_limit_override')
        try:
            if num_results_override:
                self.limit = min(self.limit, int(num_results_override))
        except:
            log(DEBUG, 'Illegal value of res_limit_override "%s" in config' % (num_results_override,))

        # Note that some parameters are processed differently here than as in begin_lookup()
        self.use_epoch_time = not extractHumanTime(qs)

        # disallow offset
        if extractOffset(qs) > 0:
            raise dnstli2.response.OffsetUnsupported

        self.aggregate = extractAggr(qs)
        self.querierId = extractId(qs)
        self.querierSWClient = extractSWClient(qs)
        self.querierVersion = extractVersion(qs)

        otherQueryParamsToLog = extractOtherQueryParamsToLog(qs)

        try:
            index, index_type, args = args.split('/', 2)
        except:
            raise dnstli2.response.BadRequest

        try:
            if index == 'rrset':
                self.query = rrsetQuery(index_type, args,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'], aggregate=self.aggregate,
                        otherQueryParamsToLog = otherQueryParamsToLog)
            elif index == 'rdata':
                self.query = rdataQuery(index_type, args,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'], aggregate=self.aggregate,
                        otherQueryParamsToLog = otherQueryParamsToLog)
            else:
                raise dnstli2.response.BadRequest

            self.row_count = 0
            self.t0 = time.time()
            self.res = self.client.lookup(self.query.qobj, self.limit)
        except dnstable.DnstableException:
            # Note that dnstable query timeouts do *not* make it here
            raise dnstli2.response.BadRequest

class rrsetQuery:
    fmt_log_type = 'rrset'

    def __init__(self, index_type, args,
                 time_first_before=None, time_first_after=None, time_last_before=None, time_last_after=None,
                 timeout=None, aggregate=None, offset=0, otherQueryParamsToLog = None, wildcard = None):
        self.format_entry = dnstli2.fmt.format_rrset
        self.log = OrderedDict()
        self.log['server'] = 'dnstli2'
        self.log['type'] = self.fmt_log_type
        cs = False

        try:
            name, rrtype, bailiwick = dnstli2.util.padded_split(args, '/', 3)

            if name:
                name = urllib.parse.unquote(name)
                if wildcard is not None and name.startswith(wildcard + '.'):
                    cs = True

            if bailiwick:
                bailiwick = urllib.parse.unquote(bailiwick)
        except:
            raise dnstli2.response.BadRequest

        if index_type == 'name':
            self.log['name'] = name
        elif index_type == 'raw':
            self.log['raw'] = name
            try:
                name = wdns.domain_to_str(binascii.unhexlify(name))
            # except TypeError:
            # except wdns.NameException:
            except:            
                raise dnstli2.response.BadRequest
        else:
            raise dnstli2.response.BadRequest

        self.qobj = dnstable.query(dnstable.RRSET, name, rrtype, bailiwick,
                time_first_before=time_first_before,
                time_first_after=time_first_after,
                time_last_before=time_last_before,
                time_last_after=time_last_after,
                timeout=timeout, aggregate=aggregate, offset=offset, case_sensitive=cs)

        if rrtype:
            self.log['rrtype'] = rrtype.upper()
        if bailiwick:
            self.log['bailiwick'] = bailiwick
        if aggregate == False:       # don't put aggr when it is the default, True
            self.log['aggr'] = aggregate
        if otherQueryParamsToLog:
            self.log['otherParams'] = otherQueryParamsToLog

class rdataQuery:
    fmt_log_type = 'rdata'

    def __init__(self, index_type, args,
                 time_first_before=None, time_first_after=None, time_last_before=None, time_last_after=None,
                 timeout=None, aggregate=None, offset=0, otherQueryParamsToLog = None):
        self.format_entry = dnstli2.fmt.format_rdata
        self.log = OrderedDict()
        self.log['server'] = 'dnstli2'
        self.log['type'] = self.fmt_log_type

        try:
            rdata, rrtype = dnstli2.util.padded_split(args, '/', 2)

            if rdata:
                rdata = urllib.parse.unquote(rdata)

            if rrtype:
                # verify rrtype is defined or our special ANY-DNSSEC type; else throw
                if rrtype.lower() != 'any-dnssec':
                    wdns.str_to_rrtype(rrtype)
        except:
            raise dnstli2.response.BadRequest

        if index_type == 'name':
            self.qobj = dnstable.query(dnstable.RDATA_NAME, rdata, rrtype,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after,
                    timeout=timeout, aggregate=aggregate, offset=offset)
            self.log['name'] = rdata
        elif index_type in ('ip', 'ip6'):
            # Jira issue DNSDB-77: restrict rrtypes allowed in rdata/ip* queries.
            # Note that we don't pass the rrtype down to dnstable.query -- it will generate its
            # own rrtype based upon the length of the IP address (4 gets A, 16 gets AAAA).
            if rrtype and rrtype.upper() not in ('A', 'AAAA', 'ANY'):
                raise dnstli2.response.BadRequest
            rdata = rdata.strip()
            rdata = rdata.replace(',', '/')
            self.qobj = dnstable.query(dnstable.RDATA_IP, rdata,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after, timeout=timeout, aggregate=aggregate, offset=offset)
            self.log['ip'] = rdata
        elif index_type == 'raw':
            self.qobj = dnstable.query(dnstable.RDATA_RAW, rdata, rrtype,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after, timeout=timeout, aggregate=aggregate, offset=offset)
            self.log['raw'] = rdata
        else:
            raise dnstli2.response.BadRequest

        if rrtype:
            self.log['rrtype'] = rrtype.upper()
        if aggregate == False:       # don't put aggr when it is the default, True
            self.log['aggr'] = aggregate
        if otherQueryParamsToLog:
            self.log['otherParams'] = otherQueryParamsToLog
